@@ -29,6 +29,7 @@
 #include <errno.h>
 
 #include <libavcodec/avcodec.h>
+#include <libavutil/stereo3d.h>
 
 #include <gst/gst.h>
 #include <gst/video/gstvideometa.h>
@@ -125,6 +126,7 @@ gst_ffmpegvidenc_base_init (GstFFMpegVidEncClass * klass)
   GstPadTemplate *srctempl = NULL, *sinktempl = NULL;
   GstCaps *srccaps = NULL, *sinkcaps = NULL;
   gchar *longname, *description;
+  const gchar *classification;
 
   in_plugin =
       (AVCodec *) g_type_get_qdata (G_OBJECT_CLASS_TYPE (klass),
@@ -134,8 +136,11 @@ gst_ffmpegvidenc_base_init (GstFFMpegVidEncClass * klass)
   /* construct the element details struct */
   longname = g_strdup_printf ("libav %s encoder", in_plugin->long_name);
   description = g_strdup_printf ("libav %s encoder", in_plugin->name);
+  classification =
+      gst_ffmpeg_codecid_is_image (in_plugin->id) ? "Codec/Encoder/Image" :
+      "Codec/Encoder/Video";
   gst_element_class_set_metadata (element_class, longname,
-      "Codec/Encoder/Video", description,
+      classification, description,
       "Wim Taymans <wim.taymans@gmail.com>, "
       "Ronald Bultje <rbultje@ronald.bitfreak.net>");
   g_free (longname);
@@ -161,6 +166,9 @@ gst_ffmpegvidenc_base_init (GstFFMpegVidEncClass * klass)
   gst_element_class_add_pad_template (element_class, srctempl);
   gst_element_class_add_pad_template (element_class, sinktempl);
 
+  gst_caps_unref (sinkcaps);
+  gst_caps_unref (srccaps);
+
   klass->in_plugin = in_plugin;
   klass->srctempl = srctempl;
   klass->sinktempl = sinktempl;
@@ -600,6 +608,31 @@ buffer_info_free (void *opaque, guint8 * data)
   g_slice_free (BufferInfo, info);
 }
 
+static enum AVStereo3DType
+stereo_gst_to_av (GstVideoMultiviewMode mview_mode)
+{
+  switch (mview_mode) {
+    case GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE:
+      return AV_STEREO3D_SIDEBYSIDE;
+    case GST_VIDEO_MULTIVIEW_MODE_TOP_BOTTOM:
+      return AV_STEREO3D_TOPBOTTOM;
+    case GST_VIDEO_MULTIVIEW_MODE_FRAME_BY_FRAME:
+      return AV_STEREO3D_FRAMESEQUENCE;
+    case GST_VIDEO_MULTIVIEW_MODE_CHECKERBOARD:
+      return AV_STEREO3D_CHECKERBOARD;
+    case GST_VIDEO_MULTIVIEW_MODE_SIDE_BY_SIDE_QUINCUNX:
+      return AV_STEREO3D_SIDEBYSIDE_QUINCUNX;
+    case GST_VIDEO_MULTIVIEW_MODE_ROW_INTERLEAVED:
+      return AV_STEREO3D_LINES;
+    case GST_VIDEO_MULTIVIEW_MODE_COLUMN_INTERLEAVED:
+      return AV_STEREO3D_COLUMNS;
+    default:
+      break;
+  }
+  GST_WARNING ("Unsupported multiview mode - no mapping in libav");
+  return AV_STEREO3D_2D;
+}
+
 static GstFlowReturn
 gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
     GstVideoCodecFrame * frame)
@@ -619,6 +652,16 @@ gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
         GST_BUFFER_FLAG_IS_SET (frame->input_buffer, GST_VIDEO_BUFFER_FLAG_TFF);
   }
 
+  if (GST_VIDEO_INFO_MULTIVIEW_MODE (info) != GST_VIDEO_MULTIVIEW_MODE_NONE) {
+    AVStereo3D *stereo = av_stereo3d_create_side_data (ffmpegenc->picture);
+    stereo->type = stereo_gst_to_av (GST_VIDEO_INFO_MULTIVIEW_MODE (info));
+
+    if (GST_VIDEO_INFO_MULTIVIEW_FLAGS (info) &
+        GST_VIDEO_MULTIVIEW_FLAGS_RIGHT_VIEW_FIRST) {
+      stereo->flags = AV_STEREO3D_FLAG_INVERT;
+    }
+  }
+
   if (GST_VIDEO_CODEC_FRAME_IS_FORCE_KEYFRAME (frame))
     ffmpegenc->picture->pict_type = AV_PICTURE_TYPE_I;
 
